\ POINTER.F     External memory allocation and management   by Tom Zimmer

.( Loading Memory Pointer...)

\ BODY          +0             +4      +8            +12
\               [ phy_pointer ][ link ][ size_bytes ][.....

VARIABLE PHEAD  PHEAD OFF

: POINTERS-INIT ( -- )
    PHEAD
    BEGIN   @ ?DUP
    WHILE   DUP CELL- OFF           \ clear each pointer in turn
    REPEAT  ;

initialization-chain chain-add pointers-init

: TRIM-POINTERS ( a1 -- a1 )
    PHEAD
    BEGIN   @ ?DUP
    WHILE   2DUP TRIM?
      IF      DUP CELL- @ ?DUP
        IF  RELEASE  THEN
      THEN
    REPEAT  DUP PHEAD FULL-TRIM ;

forget-chain chain-add trim-pointers

: .POINTERS     ( -- )
    PHEAD
    BEGIN   @ ?DUP
    WHILE   DUP CELL- BODY> >R
      CR DUP CELL+ @ 10 U,.R ."  bytes at: "
      DUP CELL- @ ?DUP
      IF      13 H.R
      ELSE    ." <unallocated>"
      THEN    ."  for: " R> .NAME
    REPEAT  ;

internal

: ?VALID_POINTER ( plink -- plink )     \ check for valid pointer sturcture
    DUP 2 cells+ @ OVER <>
    IF      cell- BODY> DUP CR h. .name
      ." is NOT a valid POINTER structure!"
      ABORT
    THEN ;

: %UnPointer    ( cfa -- )      \ deallocate pointer given the cfa
    >BODY DUP>R @ 0<>                       \ only if non-zero
    IF      R@ cell+ ?VALID_POINTER DROP
      R@ @ release            \ release the memory
      r@ off                  \ clear the memory pointer
    THEN    R>DROP ;

: %SizeOf!      ( bytes cfa -- )    \ set the size of pointer "cfa"
    >BODY cell+ ?VALID_POINTER cell+ ! ;

: %SizeOf@      ( cfa -- bytes )    \ get the size of pointer "cfa"
    >BODY cell+ ?VALID_POINTER cell+ @ ;

external

: POINTER       ( bytes -<name>- )              \ make a pointer "name"
    128 max                         \ at least 160 bytes
    CREATE 0 ,                      \ initialize to unallocated
    HERE  PHEAD @ ,  PHEAD !        \ link into chain
    ,                               \ lay in size in bytes
    PHEAD @ ,                       \ POINTER CHAIN verify word
    DOES>   dup @ 0=
      if      dup 2 cells+ @ malloc ?dup
        if      dup off   over !
        else    true abort" Failed to allocate POINTER"
        then
      then    @ ;

: UNPOINTER>    ( name -- )     \ deallocate a pointer name following
    ' STATE @
    IF POSTPONE LITERAL POSTPONE %UNPOINTER
    ELSE    %UNPOINTER
    THEN    ; IMMEDIATE

: SIZEOF!>      ( bytes | name -- ) \ set size of the following pointer
    ' STATE @
    IF POSTPONE LITERAL POSTPONE %SIZEOF!
    ELSE    %SIZEOF!
    THEN    ; IMMEDIATE

: SIZEOF@>      ( -<name>- bytes )  \ get size of the following pointer
    ' STATE @
    IF  POSTPONE LITERAL POSTPONE %SIZEOF@
    ELSE    %SIZEOF@
    THEN    ; IMMEDIATE

: NAMED-NEW$    ( -<name>- )            \ a semi-static buffer of MAXSTRING length
    MAXSTRING Pointer ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\               Dynamic String Support Words
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ The dynamic string buffer is allocated when it is first used, and automatically
\ released when the program terminates.

internal

1024 constant max-dyn-string            \ dynamic string space is this big

max-dyn-string Pointer dyn-ptr    \ the dynamic string buffer

0 value dyn-offset                      \ top of the dynamic string buffer

: DynAlloc      ( n1 -- a1 )            \ allocate space in dynamic buffer
                                        \ return a1 a buffer of n1 bytes
    dup
    max-dyn-string 4 / >    \ limit one string to 1/4 of total
    abort" Can't allocate a dynamic string this large"
    dyn-offset over + max-dyn-string >=
    if      0 to dyn-offset
    then    dyn-ptr dyn-offset + swap +to dyn-offset ;

: _new$         ( -- a1 )               \ allocate the next MAXSTRING buffer
    MAXSTRING DynAlloc ;

' _new$ is new$

module
.( OK) cr