.( Loading utilities ... )

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       User specifiable string delimiter utility
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: ,$            ( -< #text#>- )
    >in @ bl word swap >in ! 1+ c@
    word count here place here c@ 1+ allot 0 c, align ;

: .$            ( -< #text#>- )
    POSTPONE (.") ,$ ; immediate

: s$ ( -< #text#>- )
     ( -- a1 n1 )
    POSTPONE (s") ,$ ; immediate

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       Multiple directory path search capability for file open
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

INTERNAL

create current-dir$  maxstring allot
    create &fpath    MAXSTRING allot     \ a static forth path buffer
           &fpath    off
  variable &linenum  &linenum off

2variable path-source

&fpath value path-ptr                   \ initialize the path buffer pointer

EXTERNAL

: next-path"    ( -- a1 n1 )            \ get the next path from dir list
    path-source 2@ 2dup ';' scan  2dup 1 /string path-source 2!
    nip - ;

: first-path"   ( -- a1 n1 )            \ get the first forth directory path
    path-ptr count path-source 2!
    next-path" ;

: "fpath+       ( a1 n1 -- )            \ append a directory to forth path
    2dup + 1- c@ '/' =              \ end in '/'?
    if      1- 0max                 \ if so, delete it
    then    first-path"                     \ get first path
    begin   dup>r 2over compare dup r> and  \ check it
    while   drop
      next-path"                      \ and remaining paths
    repeat  0=              \ -- f1=true if already in list
    if      2drop
    else    path-ptr dup c@ if ?+; else drop then
            path-ptr +place
    then    ;

: .fpath        ( -- )          \ display the forth directory search path list
    path-ptr count
    begin   ?dup
    while   2dup ';' scan 2dup 2>r nip - ( dup 1+ ?cr ) type
      2r> 1 /string dup
      if ." ;" then
    repeat  drop ;

: fpath+        ( -<directory>- )       \ append a directory to forth path
    bl word count "fpath+ ;

INTERNAL

create open-save$ MAXSTRING allot       \ buffer to save the file being opened
create open-path$ MAXSTRING allot
                                        \ f1=FALSE=success, TRUE=failed
: n"open        ( a1 n1 -- handle f1 )          \ open file a1,n1 with path search
    open-save$ place                \ save filename for later
    open-save$ count _"open dup     \ if we couldn't open the file
    if open-save$ 1+ c@ '/' <>      \ not if  first is '/'
      if 2drop                           \ discard _"open results
        first-path"
        begin dup>r
          open-path$ place        \ first path
          open-path$ c@           \ append if path is there
          if open-path$ ?+/  \ plus '/'
          then
          open-save$ count
          open-path$ +place               \ append name
          open-path$ count _"open dup     \ open it
          r> and
        while 2drop next-path"
        repeat
      then
    else open-save$ 1+ c@ '/' <>   \ not if  first is '/'
      if current-dir$ count open-path$ place
        open-path$ ?+/
        open-save$ count open-path$  +place
      else    open-save$   count open-path$   place
      then
    then    ;               \ return n2=handle, f1=false if found

' n"open is "open       \ link multi-path open word into system

EXTERNAL

: "path-file    ( a1 n1 -- a2 n2 f1 )   \ find file a1,n1 return full path
                                        \ a2,n2 and f1=false, succeeded
                                        \ else return a1,n1 and f1=true, failed
                2dup "open 0=
                if      close-file drop
                        2drop
                        open-path$ count false
                else    drop true
                then    ;

: "file-clip"   { adr len limit \ temp$ pre -- 'adr 'len }
                                                        \ clip filename to limit
                MAX-PATH LocalAlloc: temp$
                limit 20 max to limit                   \ must be at east 16
                limit 20 - 2 / 6 + to pre               \ balance start and end
                len limit >
                if      adr pre 3 -     temp$  place    \ lay in first 5 chars
                        s" ..."         temp$ +place    \ append some dots
                        adr len dup limit pre - - /string \ clip to last part
                                        temp$ +place    \ of name and lay in
                                        temp$  count
                else    adr len                         \ no need to clip file
                then    ;
comment:
: .platform     ( -- )
                cr ." Platform: "
                call GetVersion dup 0x7FFFFFFF and swap 0x80000000 and
                if      dup word-split drop 0x100 mod 4 <
                        if      ." Win32s on Windows"
                        else    ." Windows 95"
                        then
                else    ." Windows NT"
                then    word-split swap
                ."  Version: " 0x100 /mod swap 1 h.r ." ." .
                ."  Build: " . ;
comment;

: where         ( -<name>- )    \ tell me WHERE the source for a word is
    '
    loadfile
    begin @ ?dup
    while 2dup >
      if ." was loaded from " swap
        >name 1+ @ ?dup
        if  swap  cell+ count type ."  line " .
        else  drop ." CONSOLE"  then cr
        exit
      then
    repeat ;

: .loaded
    loadfile
    begin @ ?dup
    while dup cell+ count type space
    repeat cr ;

: "loaded?      { adr len \ temp$ -- f1 }       \ f1 = true if file has been loaded
                max-path LocalAlloc: temp$      \ allocate local string buffer
                adr len upper
                adr len "to-pathend" to len to adr      \ strip off path
                loadfile @      \ top of the file loaded chain
                begin   ?dup    \ for as long as we aren't at the end
                while   dup cell+ count "to-pathend" temp$ place
                        temp$ ?defext
                        temp$ count upper
                        temp$ count adr len compare 0=
                        if      drop true
                                EXIT            \ exit with true on stack
                        then
                        dup @
                        if      \ if no code compiled, then discard filename
                                dup>r @ dup cell+ count + r> =
                                if      @
                                then
                        else    @
                        then
                repeat  false ;

: NEEDS         ( -<name>- ) \ conditionally load file "name" if not loaded
                >in @ >r
                bl word ?defext                 \ add file ext if needed
                pocket count "loaded? 0=        \ if file isn't loaded
                if      r@ >in !
                        fload                   \ then loadit
                then    r>drop ;

: \LOADED-      ( -<name>- ) \ if the following file IS NOT LOADED interpret line
                >in @ >r
                bl word ?defext
                pocket count "loaded? 0=
                if      interpret
                else    postpone \
                then    r>drop ;

: \LOADED       ( -<name>- ) \ if the following file IS LOADED interpret line
                >in @ >r
                bl word ?defext
                pocket count "loaded?
                if      interpret
                else    postpone \
                then    r>drop ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       Compiler utilities
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: anew          ( -<name>- )    \ define a new marker
                >in @ defined
                if      execute  \ July 25th, 1997 tjz removed NIP
                else    drop
                then    >in ! marker ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       Utility to allow loading a file starting at a specified line number
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: $fload
    count included ;

: #fload        ( n1 -<name>- )         \ load file "name" from line n1, 1 based
                start-line !                    \ set start line
                bl word $fload ;                \ do the load

: lineload      ( n1 -- )               \ load the current file from line n1
                start-line !
                cur-file $fload ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       Display the amount of used and available program memory
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: freedic sp@ here - ;

: .free         ( -- )
    base @ decimal
\    cr ." Application address:   "   &origin             h.8  ." h"
\    cr ."               Total: " &app-size @               10 u,.r
    cr ."                Used: " here                      10 u,.r
    cr ."                Free: " freedic                    10 u,.r
\    cr ." Malloc/Pointer  Mem: " tot-malloc                10 u,.r
    base ! ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       A simple error number extention to error handling
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: ?error        ( f1 n1 -- )    \ abort with error code n1 if f1=true
                swap
                if      throw
                else    drop
                then    ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       ANSI Save and Restore Input Functions
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: save-input    ( -- xxx 7 )
                loadfile @ cell+
                ?loading @
                loadline @
                >in @
                source-id
                (source) 2@
                7 ;

: restore-input ( xxx 7 -- f1 )
                drop
                (source) 2!
                to source-id
                >in !
                loadline !
                ?loading !
                align linkfile 0 ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       Compile time stack depth checking
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ synonym checkstack nostack1
INTERNAL

0 value olddepth

: stack-empty?  ( -- )
                depth abort" The stack should have been empty here!" ;

: _stack-check  ( -- )
                ?loading @ 0=           \ if we are not loading
                state @ or              \ or we are in compile state,
                                        \ then don't check stack depth change
                olddepth 0< or ?exit    \ or if olddepth is below zero
                                        \ or if assembling
                context @ [ ' assembler vcfa>voc ] literal = ?exit
                depth olddepth >        \ if stack depth has increased
                if                      \ then warn of extra item on stack
                        cr ." Stack depth increased in file: "
                        loadfile @ cell+ count type
                        ."  at line: " base @ decimal loadline @ . base !
                        ." Stack: " .s cr
                then    depth to olddepth ;
EXTERNAL

: nostack       ( -- )
                -1 to olddepth ;

nostack ' _stack-check is stack-check

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\      Random number generator for Win32Forth
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

3141592 value SEED1
6535897 value SEED2
9323846 value SEED3

: RANDOM        ( n1 -- n2 )    \ get a pseudo random number between 0 and n1 as n2
                dup 0= if 1+ then
                SEED1 177 /MOD  2*    SWAP  171 *  SWAP  - DUP  to SEED1
                SEED2 176 /MOD  35 *  SWAP  172 *  SWAP  - DUP  to SEED2
                SEED3 178 /MOD  63 *  SWAP  170 *  SWAP  - DUP  to SEED3
                +  + SWAP  MOD  ;
comment:
: RANDOM-INIT   ( -- )          \ initialize the random number generator
\                get-local-time
                time-buf 3 cells + @ to SEED1
                time-buf 2 cells + @ to SEED2
                time-buf 1 cells + @ to SEED3 ;

INITIALIZATION-CHAIN CHAIN-ADD RANDOM-INIT      \ randomize at boot time
comment;
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       Utility to type a file to the console
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

variable cur-line

: "ftype        { \ locHdl typ$ -<name>- }   \ type file "name" to the console
                max-path LocalAlloc: typ$
                "open abort" Couldn't open file!"
                to locHdl
                cur-line off
                cr ." Typing file: " open-path$ count type cr
                begin   typ$ dup MAXCOUNTED locHdl read-line
                        abort" Read Error"
                        ( nuf? 0= and )
                while   type cr
                        ( 10 ms )
                repeat  2drop
                locHdl close-file drop ;

: ftype         ( -<filename>- )
                bl word count "ftype ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\       An addition to CASE OF ENDOF ENDCASE, to allow testing ranges
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: _of-range     ( n1 n2 n3 -- n1 f1 )
                2 pick -rot between ;

: of-range      ( n1 n2 n3 -- n1 )      \ extention to CASE for a range
                ?comp POSTPONE _of-range POSTPONE ?branch here 4 ; immediate

: 2literal      ( d1 -- )
                swap POSTPONE LITERAL POSTPONE LITERAL ; immediate

\ --------------------------------------
\   extended strings
\ --------------------------------------

\ use ^J or \x0a or \n for a newline, \t for tab, \0 for - you know

: xstr>txt ( ad1 n ad2 -- )
    tuck >r
    0 ?do 1+  over i + count 
      case
        '\' of count
               case
                 't' of drop 2 9 endof
                 'n' of drop 2 10 endof
                 'x' of 4 swap  count
                        dup '9' > if 0xdf and 7 - then '0' -  16 *
                        swap c@
                        dup '9' > if 0xdf and 7 - then '0' -  + endof
                 '0' of drop 2 0 endof
               endcase endof
        '^' of c@  31 and  2 swap endof
        nip 1 tuck
      endcase
      2 pick c!
    +loop nip  0 over 1+ c!
    r@ -  r> c! ;

:noname [char] " word count  new$ dup>r  xstr>txt  r> count ;
:noname compile (s")  [char] " word count  here xstr>txt
    here c@ 2 + aligned allot ;
interpret/compile: i"

only forth also definitions
.( OK) cr